package Oggify;

# Oggify -- Tool to assist with bulk FLAC encoding
# Copyright (c) 2005,2006 Scott Paul Robertson (spr@mahonri5.net)
#
# This is part of Oggify (http://spr.mahonri5.net/oggify/)
#
# Oggify is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version
# 
# Oggify is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with Oggify; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

use strict;
use warnings;

require Exporter;
our @ISA = qw(Exporter);
our @EXPORT_OK = qw($Flac_Tree_Head $Output_Tree_Head $Mode $Not_Mode
                    $Quality $Verbose $Nice_Value $Enc_Type $Symlink
                    $VERSION $Tmp_File $Force %Encode @Purge
                    %Reencode @Lim_Purge process ogg_enc mp3_enc
                    retag_file purge cleanup sigdie_handler);
our $VERSION = '1.0.2';

use Audio::TagLib;
use File::Find;
use File::Basename;

# Exportable Global variables {{{1
our $Flac_Tree_Head = '';       # Root of the directory tree with the flac's
our $Output_Tree_Head = '';     # Root of the directory tree to output to
our $Mode = ".ogg";             # Mode we will be encoding in
our $Not_Mode = ".mp3";         # I know it's lame.
our $Quality = 5;               # Quality to encode at
our $Verbose = 0;               # Verbosity.
our $Nice_Value = 10;           # Default value to nice processes at
our $Enc_Type = "vbr";          # Mp3 encoding mode
our $Symlink = 0;               # Do we follow symlinks?
our $Hidden_Files = 0;          # Do we look at hidden files?
our $Tmp_File = 0;              # The tmp file used for encoding
our $Force = 0;                 # Force purging regardless of contents

our %Encode = ();               # flac_file -> output_file
our @Purge = ();                # Files to delete
our %Reencode = ();             # flac_file -> output_file
our @Lim_Purge = ();            # Files to delete most times
# }}}1

# Local globals {{{1
my %flac_dirs = ();
my %flac_files = ();
my $purge_ok = 1;
my %mp3_quality_conversion = (
    cbr => [
        '--preset cbr 64',
        '--preset cbr 64',
        '--preset cbr 96',
        '--preset cbr 128',
        '--preset cbr 160',
        '--preset cbr 192',
        '--preset cbr 256',
        '--preset insane',
        '--preset insane',
        '--preset insane',
        '--preset insane',
    ],
    vbr => [
        '--preset medium',
        '--preset medium',
        '--preset medium',
        '--preset standard',
        '--preset standard',
        '--preset standard',
        '--preset extreme',
        '--preset extreme',
        '--preset extreme',
        '--preset extreme',
        '--preset insane',
    ],
    abr => [
        '--preset 64',
        '--preset 64',
        '--preset 96',
        '--preset 128',
        '--preset 160',
        '--preset 192',
        '--preset 256',
        '--preset 320',
        '--preset 320',
        '--preset 320',
        '--preset insane',
    ],
);
# }}}1

# Processing {{{1
sub process {

    my %options = ('wanted' => \&_flac_wanted);
    $options{'follow'} = 1 if ($Symlink);

    find(\%options, ($Flac_Tree_Head));

    if (! -e $Output_Tree_Head){
        my $temp = $Output_Tree_Head;
        $temp = cleanstring($temp);
        `mkdir -p $temp`;
    }
    $options{'preprocess'} = \&_out_pre;
    $options{'wanted'} = \&_out_wanted;
    find(\%options, ($Output_Tree_Head));

    foreach my $key (keys %flac_files) {
        my ($filename, $path, $suffix) = fileparse($key, qr/\.[^.]+/);
        my $encode_file = $Output_Tree_Head . $path . $filename . $Mode;
        $Encode{$flac_files{$key}} = $encode_file;
    }

    # TODO: take the purge check out of purge and put it here. Also allow
    # for the passing of a 'force' flag to really really clean up
    # unmatching entries

    return;
} #}}}1

# Processing Functions {{{1
sub _out_pre {
    my @ret = ();
    foreach my $item (@_) {
        if (-d $item) {
            push @ret, $item;
        }
        elsif ($item =~ /\.mp3$|\.ogg$/) {
            push @ret, $item;
        }
    }

    return @ret;
}

sub _flac_wanted {
    if (-d $_) {
        my $trimmed = substr $File::Find::name, (length $Flac_Tree_Head);
        $flac_dirs{$trimmed} = $File::Find::name;
    }
    elsif ($_ =~ /\.flac$/) {
        my $trimmed = substr $File::Find::name, (length $Flac_Tree_Head);
        $flac_files{$trimmed} = $File::Find::name;
    }
}

sub _out_wanted {
    my $trimmed = substr $File::Find::name, (length $Output_Tree_Head);
    if (-d $_) {
        if (exists $flac_dirs{$trimmed}) {
            delete $flac_dirs{$trimmed};
        }
        elsif ($trimmed ne '') {
            push @Purge, $File::Find::name;
        }
    }
    else {
        my ($filename, $path, $suffix) = fileparse($trimmed, qr/\.[^.]+/);
        my $eq_flac_file = $path . $filename . '.flac';
        if (exists $flac_files{$eq_flac_file}) {
            if ($suffix eq $Mode) {
                $Reencode{$flac_files{$eq_flac_file}} = $File::Find::name;
                delete $flac_files{$eq_flac_file};
            }
            else {
                push @Lim_Purge, $File::Find::name;
            }
        }
        else {
            push @Purge, $File::Find::name;
        }
    }
}
# }}}1

# Actions {{{1
# Ogg {{{2
sub ogg_enc {
    my ($src_file, $out_file) = @_;
    $Tmp_File = "$out_file.tmp.$$";
    my $c_src = cleanstring($src_file);
    my $c_out = cleanstring($out_file);
    my $c_tmp_file = cleanstring($Tmp_File);

    my $out_dir = dirname($c_out);

    `mkdir -p "$out_dir"`;

    my $exec = "oggenc -Q -q $Quality -o \"$c_tmp_file\" \"$c_src\"";
    if ($Nice_Value) {
        $exec = "nice -n $Nice_Value " . $exec;
    }

    print "$exec\n" if ($Verbose > 1);
    if ($Verbose < 1) {
        `$exec`;
    }
    else {
        print `$exec`;
    }

    `mv "$c_tmp_file" "$c_out"`;
}
#}}}2

# Mp3 {{{2
sub mp3_enc {
    my ($src_file, $out_file) = @_;
    $Tmp_File = "$out_file.tmp.$$";
    my $c_src = cleanstring($src_file);
    my $c_out = cleanstring($out_file);
    my $c_tmp_file = cleanstring($Tmp_File);

    my $out_dir = dirname($c_out);

    `mkdir -p "$out_dir"`;

    my $enc_type_use = $mp3_quality_conversion{$Enc_Type}[$Quality];

    my $exec = "flac --totally-silent -d -c \"$c_src\" |";

    if ($Nice_Value) {
        $exec .= " nice -n $Nice_Value";
        $exec = "nice -n $Nice_Value " . $exec;
    }

    $exec .= " lame -S $enc_type_use";
    $exec .= " - \"$c_tmp_file\"";

    print "$exec\n" if ($Verbose > 1);
    if ($Verbose < 1) {
        `$exec 2>&1`;
    }
    else {
        print `$exec`;
    }

    `mv "$c_tmp_file" "$c_out"`;

    print "tagging with TagLib\n" if ($Verbose > 1);

    retag_file($src_file, $out_file);
}
#}}}2

# Re-Tag {{{2
sub retag_file {
    my ($src_file, $out_file) = @_;
    my ($out_f, $out_tag);

    my $flac_f = Audio::TagLib::FLAC::File->new($src_file);
    my $flac_tag = $flac_f->tag();

    if ($Mode eq ".mp3") {
        $out_f = Audio::TagLib::MPEG::File->new($out_file);
    }
    elsif ($Mode eq ".ogg") {
        $out_f = Audio::TagLib::Vorbis::File->new($out_file);
    }
    else {
        $out_f = Audio::TagLib::FileRef->new($out_file);
    }
    $out_tag = $out_f->tag();

    Audio::TagLib::Tag->duplicate($flac_tag, $out_tag);

    if (!$out_f->save()) {
        print "Unable to save tags on $out_file!\n";
    }
    
    # Fix ID3v2 Genre
    if ($Mode eq ".mp3" && $flac_tag->genre()->toCString) {
        my $out_f = Audio::TagLib::MPEG::File->new($out_file);
        _fix_genre_tag($out_f);
        if (!$out_f->save()) {
            print "Unable to save tags on $out_file!\n";
        }
    }
}

# This ensures ID3v2 Genre's are strings, not numbers.
# Tagging with numbers complies with the standard, but many players
# do not follow it correctly (iTunes/iPod), so we tag with the string.
# TagLib will change to this behavior in v1.5.
sub _fix_genre_tag {
    my $file = shift;
    my $tag = $file->ID3v2Tag();
    my $i = $tag->frameListMap();
    tie my %i, ref($i), $i;

    my $l = $i{Audio::TagLib::ByteVector->new("TCON")};
    my $genre = $tag->genre();
    $l->begin()->data()->setText($genre);
}
#}}}2

# Purge {{{2
sub purge {
    my $to_kill = shift;

    $purge_ok = 1;
    if (-d $to_kill) {
        my %options = ('wanted' => \&_purge_wanted, 'bydepth' => 1, 'follow' => 1);
        find(\%options, ($to_kill));
    }

    $to_kill = cleanstring($to_kill);
    `rm -rf "$to_kill"` if ($purge_ok || $Force);
}

sub _purge_wanted {
    if ( -e $_ && ! -d $_ ) {
        $purge_ok = 0;
    }
} # }}}2
#}}}1

# Cleanup and Signal Handling {{{1
sub cleanup {
    if ($Tmp_File && -e "$Tmp_File") {
        unlink "$Tmp_File";
    }
}

sub sigdie_handler {
    local $SIG{INT} = 'IGNORE';
    kill INT => -$$;
    cleanup();
    exit(-1);
} # }}}1

sub cleanstring {
    my ($dirty) = @_;
    my $clean = $dirty;
    $clean =~ s/\\/\\\\/g;
    $clean =~ s/"/\\"/g;
    $clean =~ s/\$/\\\$/g;
    $clean =~ s/`/\\`/g;
    return $clean;
}

1;
