#!/usr/bin/perl -w

# oggify.pl -- Command-line interface for Oggify
# Copyright (c) 2005,2006 Scott Paul Robertson (spr@mahonri5.net)
#
# This is part of Oggify (http://spr.mahonri5.net/oggify/)
#
# Oggify is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version
# 
# Oggify is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with Oggify; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

use strict;
use File::Basename;
use Getopt::Long;
Getopt::Long::Configure ("bundling");
use FindBin;
use lib "$FindBin::RealBin";
use Oggify qw($Flac_Tree_Head $Output_Tree_Head $Mode $Not_Mode $Quality
            $Verbose $Nice_Value $Enc_Type $Symlink $Force
            $Tmp_File %Encode @Purge %Reencode @Lim_Purge);

my $version = $Oggify::VERSION;

# Help message {{{1
my $help_message = <<"END_BASIC_HELP";
oggify.pl ($version) -- Easily convert flac files to a more common format.
usage: oggify.pl [options] <source> <destination>
<source> is a directory or flac file, <destination> is the output directory.
Will recurse through <source> and encode and output to <destination>.
Options:
-h (--help) -- This message.
-H (--full-help) -- Verbose help, this examples.
-q (--quality) <number> -- sets the quality to <number> (as oggenc, default 5)
-v (--verbose) -- Increases output details.
-t (--test) -- Print out what actions will be done.
-n (--nice) <value> -- nices the encodes at value (default 10).
-p (--purge) -- remove unmatched items from the destination.
-r (--refresh) -- re-encode files that are older than the source.
-R (--retag) -- Disables other actions, retags files older than the source.
-f (--force) -- Disables checking for refresh, retag, and purge.
-c (--clean) -- Removes files in the wrong format.
-L (--follow-symlinks) -- oggify will follow symlinks.
--mp3 -- Generate mp3's using lame instead of ogg's.
--enc-type (cbr|vbr|abr) -- (mp3 only) sets lame's encoding (default vbr).
END_BASIC_HELP
# }}}1

# Verbose Help {{{1
my $verbose_help_message = << "END_VERB_HELP";
oggify.pl -- Converts a directory tree of flacs into a similar directory tree of
oggs or mp3s. Will also handle single files.
Version: $version
Use: oggify.pl [options] <source> <destination>
where <source> is either a flac file or directory, and <destination> is a
directory.
Oggify will walk the directory structure found in <source> and create an equiv-
alent one in <destination>, and encode any flac files found into mp3's or ogg's
oggify.pl is licensed under the GPL, please read the source code for more info.

--------------------------------------------------------------------------------

Options and explanations:
-h (--help) -- Prints a simple help
-H (--full-help) -- Prints this help
-q <number> or --quality <number>
	Sets the quality level oggify will use to encode your files.
	Values: 0 - 10. 0 is the lowest, 10 the highest.
	If encoding to ogg, the quality value is the same as the -q switch for
	oggenc, please see the oggenc(1) manpage
	If encoding to mp3, the quality value is determined via a lookup table
	defined as follows:
		----------------------------------------
		| Mode |  quality  | lame setting      |
		----------------------------------------
		| cbr  |   0 - 1   | --preset cbr 64   |
		| cbr  |     2     | --preset cbr 96   |
		| cbr  |     3     | --preset cbr 128  |
		| cbr  |     4     | --preset cbr 160  |
		| cbr  |     5     | --preset cbr 192  |
		| cbr  |     6     | --preset cbr 256  |
		| cbr  |     7     | --preset insane   |
		| vbr  |   0 - 2   | --preset medium   |
		| vbr  |   3 - 5   | --preset standard |
		| vbr  |   6 - 9   | --preset extreme  |
		| abr  |   2 - 6   | same as cbr       |
		| abr  |   7 - 9   | --preset abr 320  |
		| any  |    10     | --preset insane   |
		----------------------------------------

	vbr is the default when encoding mp3's.
	for more information on the lame setting please see the lame(1)
	manpage.
-v or --verbose
	Increases the details that are printed concerning the encoding.
-t or --test
	Prints out what oggify.pl would do.
-n <value> or --nice <value>
	Nices the encoding called by oggify.pl.
	<value> can be any value nice (see nice(1) manpage) will accept.
	Ensure that the user running oggify.pl can nice to that value.
-p or --purge
	Any mp3's or ogg's that are found in the destination tree that are not
	in the source tree will be removed if purge is enabled. Any empty
	directories that do not match will be removed. To remove all unmatched
	directories pass the 'force' option.
-r or --refresh
	Any output files that are older than the flac's mtime will be
	re-encoded based on the settings you supply. To re-encode all files
	pass the 'force' option.
-R or --retag
	Disables other actions and retags files with older mtimes than the
	source file.  To retag all files, pass the 'force' option.
	Note this only affects files that would normally match:
	IE: --mp3 must be passed to retag mp3's.
-f or --force
        Ignore age of flac files when running with refresh or retag. This means
        the entire tree will be re-encoded or retaged. Also ignore whether a
	directory is empty with purge.
-c or --clean
	If a file is found in the output tree that matches one in the source,
	but has the wrong encoding (.mp3 when doing oggs) it will delete the
	file. (A file in the correct format will be made regardless)
-L or --follow-symlinks
	Tells oggify to pay attention to symlinks, will go into directories
	that are symlinks, and will take symlinks into consideration when
	determining the need to encode.
--mp3
	Sets oggify.pl to encode to mp3
--enc-type <cbr|vbr|abr> (default: vbr)
	Sets the encode type for mp3's to cbr (constant bit rate) or
	abr (average bit rate). If unspecified oggifly will use vbr
	(variable bit rate). See lame(1) manpage for more info
Examples:
	oggify.pl /mnt/flac /home/music
	Transverses /mnt/flac, creating a similar directory structure in
	/home/music, and will create any files missing from /home/music
	that exist as flac's in /mnt/flac. Those that already exist will
	be updated if mtimes differ

	oggify.pl --mp3 --enc-type vbr flac/modern music
	Author's standard method of use
	Encoding will make vbr mp3's with the --preset standard switch
	passed to lame

        -oggify.pl -q 7 \"/flac/weezer/maladriot/keep fishin.flac\" /tmp
	-Will create a quality 7 ogg file: /tmp/keep fishin.ogg

	oggify.pl -vrtq 7 --mp3 --enc-type cbr /home/flac /home/music
	verbose, refresh, test, quality 7 mp3's using cbr encoding

	oggify.pl -q 7 /home/flac /home/music/
	Generates oggs for every flac in /home/flac at quality 7

	-oggify.pl -q 6 /flacs/something.flac /outdir/this/
	-encodes /flacs/something.flac and outputs to /outdir/this/something.ogg
Quirks/Requirements:
	lame, oggenc, flac, and mkdir must all be in your path.
END_VERB_HELP
# }}}1

# Global Variables Specific to oggify.pl {{{1

my $truth = 1;      # whether we're doing this for real
my $encode = 1;     # are we encoding
my $refresh = 0;    # Whether we'll re-encode files if out of date
my $purge = 0;      # Whether to remove unmatched items
my $retag = 0;      # Are we retagging the tree?
my $fix = 0;        # Do we encode files over to the correct format?
my $debug = 0;      # debug level print outs
my $need_help = 0;
my $need_more_help = 0;

# }}}1

# Arg settings, checking {{{1
my $result = GetOptions (
        "help|h"            => \$need_help,
        "full-help|H"       => \$need_more_help,
        "quality|q=i"       => \$Quality,
        "verbose|v+"        => \$Verbose,
        "version"           => sub { print "Oggify version $version\n"; exit; },
        "silent|s+"         => sub { $Verbose = -1 },
        "test|t"            => sub { $truth = 0 },
        "nice|n=i"          => \$Nice_Value,
        "purge|p"           => \$purge,
        "refresh|r"         => \$refresh,
        "retag|R"           => \$retag,
        "clean|c"           => \$fix,
        "force|f"           => \$Force,
        "follow-symlinks|L" => \$Symlink,
        "mp3"               => sub { $Mode = ".mp3"; $Not_Mode = ".ogg"; },
        "enc-type=s"        => \$Enc_Type);

$Flac_Tree_Head = $ARGV[-2];
$Output_Tree_Head = $ARGV[-1];

if (!$result) {
    print "Need options!\n";
    print $help_message;
    exit -1;
}

if ($need_help) {
    print $help_message;
    exit;
}

if ($need_more_help) {
    print $verbose_help_message;
    exit;
}

verify_args();

if (rindex($Flac_Tree_Head, '/') == (length $Flac_Tree_Head) - 1) {
    chop $Flac_Tree_Head;
}
if (rindex($Output_Tree_Head, '/') == (length $Output_Tree_Head) - 1) {
    chop $Output_Tree_Head;
}

# }}}1

if ($Verbose > 1) {
    print "Starting at: $Flac_Tree_Head.\n";
    print "Outputing to: $Output_Tree_Head.\n";
}

# Setup signal handler
setpgrp(0,0);
$SIG{QUIT} = \&Oggify::sigdie_handler;
$SIG{INT} = \&Oggify::sigdie_handler;

# Run! {{{1
if (-d $Flac_Tree_Head) {
    Oggify::process();
}
elsif (-f $Flac_Tree_Head) {
    my $outfile = $Output_Tree_Head . '/' . basename($Flac_Tree_Head, ".flac");
    $Encode{$Flac_Tree_Head} = $outfile . '.mp3';
}

if ($encode && !$retag) {
    if ($fix) {
        print "Cleanup...\n" if ($Verbose >= 0);
        foreach my $item (sort @Lim_Purge) {
            print "Deleting $item\n" if ($Verbose >= 0);
            Oggify::purge($item) unless not $truth;
        }
    }

    print "Encoding...\n" if ($Verbose >= 0);
    foreach my $flac (sort keys %Encode) {
        print "Encoding $flac to $Encode{$flac}\n" if ($Verbose >= 0);
        if ($Mode eq ".ogg") {
            Oggify::ogg_enc($flac, $Encode{$flac}) unless not $truth;
        }
        else {
            Oggify::mp3_enc($flac, $Encode{$flac}) unless not $truth;
        }
    }
}

if ($purge && !$retag) {
    print "Purging...\n" if ($Verbose >= 0);
    foreach my $item (sort @Purge) {
        print "Deleting $item\n" if ($Verbose >= 0);
        Oggify::purge($item) unless not $truth;
    }
}

if ($retag) {       # Only if the flac file is newer
    print "Retagging...\n" if ($Verbose >= 0);
    foreach my $flac (sort keys %Reencode) {
        my $flac_mtime = (stat($flac))[9];
        my $out_mtime = (stat($Reencode{$flac}))[9];
        if ($Force || $flac_mtime > $out_mtime) {
            print "Retag $Reencode{$flac} with $flac\n" if ($Verbose >= 0);
            Oggify::retag_file($flac, $Reencode{$flac}) unless not $truth;
        }
    }
}

if ($refresh && !$retag) {    # Only if the flac file is newer
    print "Reencoding...\n" if ($Verbose >= 0);
    foreach my $flac (sort keys %Reencode) {
        my $flac_mtime = (stat($flac))[9];
        my $out_mtime = (stat($Reencode{$flac}))[9];
        if ($Force || $flac_mtime > $out_mtime) {
            print "Reencoding $flac to $Reencode{$flac}\n" if ($Verbose >= 0);
            if ($Mode eq ".ogg") {
                Oggify::ogg_enc($flac, $Reencode{$flac}) unless not $truth;
            }
            else {
                Oggify::mp3_enc($flac, $Reencode{$flac}) unless not $truth;
            }
        }
    }
}
#}}}1

Oggify::cleanup();
exit(0);

# function verify_args {{{1
sub verify_args {

    if (! $Flac_Tree_Head) {
        print "Must specify flac directory!\n";
        print $help_message;
        exit(-1);
    }

    if (! $Output_Tree_Head) {
        print "Must specify out directory!\n";
        print $help_message;
        exit(-1);
    }

    if (! -e $Flac_Tree_Head) {
        print "Source of Flacs does not exist!\n";
        print $help_message;
        exit(-1);
    }

    if ($Quality < 0 || $Quality > 10) {
        print "quality must be between 0 and 10, inclusive\n";
        print $help_message;
        exit(-1);
    }

    if ($Nice_Value > 20 || $Nice_Value < -20) {
        print "Nice values needs to be between -20 and 20, inclusive\n";
        print $help_message;
        exit(-1);
    }

    if ($Enc_Type ne "vbr" && $Enc_Type ne "cbr" && $Enc_Type ne "abr") {
        print "enc-type must be one of: vbr, cbr, abr\n";
        print $help_message;
        exit(-1);
    }

    my $prog;
    if ($Mode eq "ogg") {
        $prog = `which oggenc`;
        chomp $prog;
        if (! -e $prog) {
            print "oggenc must be in path!\n";
            print $help_message;
            exit(-1);
        }
    }
    else {
        $prog = `which lame`;
        chomp $prog;
        if (! -e $prog) {
            print "lame must be in path!\n";
            print $help_message;
            exit(-1);
        }
    }

    my @progs = ("flac", "metaflac", "mkdir");
    foreach $prog (@progs) {
        my $path = `which $prog`;
        chomp $path;
        if (! -e $path) {
            print "$prog must be in path!\n";
            print $help_message;
            exit(-1);
        }
    }

} # }}}1

# vim: sw=4 sts=4 expandtab
